import importlib.resources as lib_resources
import json
import sys
from json import JSONDecodeError
from typing import List, Dict

import click
import questionary
from jsonschema import validate, ValidationError
from questionary import Style

from src.core.exceptions import (
    InvalidTemplateException,
    CommandNotFoundException,
    ActionNotFoundException,
    ShowHelpException,
    PhoenixException,
)
from src.core.git import require_git_repo
from src.core.models import Execution
from src.core.phoenix import get_template
from src.core.rules import fire_rules
from . import resources

PHOENIX_SCHEMA = json.loads(
    lib_resources.read_text(resources, "phoenix-schema.json")
)
CUSTOM_STYLE = Style(
    [
        ("qmark", "fg:#673ab7 bold"),  # token in front of the question
        ("question", "fg:#673ab7 bold"),  # question text
        (
            "answer",
            "fg:#f44336 bold",
        ),  # submitted answer text behind the question
        (
            "pointer",
            "fg:#ff0000 bold",
        ),  # pointer used in select and checkbox prompts
        (
            "highlighted",
            "fg:#673ab7 bold",
        ),  # pointed-at choice in select and checkbox prompts
        ("selected", "fg:#cc5454"),  # style for a selected item of a checkbox
        ("separator", "fg:#cc5454"),  # separator in lists
        (
            "instruction",
            "",
        ),  # user instructions for select, rawselect, checkbox
        ("text", ""),  # plain text
        (
            "disabled",
            "fg:#858585 italic",
        ),  # disabled choices for select and checkbox prompts
    ]
)


def _is_argument(arg):
    return arg.startswith("-")


def _get_template():
    template_path = get_template()

    with open(template_path, encoding="utf-8") as template_file:
        try:
            return json.load(template_file)
        except JSONDecodeError as e:
            raise InvalidTemplateException(
                f"{e.msg} - Error at line:"
                f" {e.lineno} "
                f"and "
                f"column: {e.colno}"
            )


def _validate_if_init(args, template):
    if "init" == args[0] and {} != template.get("init"):
        raise InvalidTemplateException()


def _get_commands(template) -> List[str]:
    choices = []

    if {} != template.get("init"):
        choices.append("init")

    choices.extend(command.get("name") for command in template.get("commands"))

    return choices


def _get_command_by_template(template) -> Dict:
    command = questionary.form(
        command_name=questionary.select(
            "Choose one of the commands below",
            choices=_get_commands(template),
            style=CUSTOM_STYLE,
        )
    ).ask()

    command = command.get("command_name")

    return _get_command_by_name(command, template)


def _get_command_by_name(command, template):
    return next(
        filter(
            lambda x: x.get("name") == command,
            template.get("commands"),
        ),
        None,
    )


def _get_actions(command) -> List[str]:
    return [action.get("name") for action in command.get("actions")]


# def _get_actions_alias(command) -> List[str]:
#     return [action.get("alias") for action in command.get("actions")]


def _get_action_by_command(command) -> Dict:
    action = questionary.form(
        action_name=questionary.select(
            "Choose one of the actions below",
            choices=_get_actions(command),
            style=CUSTOM_STYLE,
        )
    ).ask()

    action = action.get("action_name")

    return _get_action_by_name(action, command)


def _get_action_by_name(action, command):
    return next(
        filter(
            lambda x: x.get("name") == action,
            command.get("actions"),
        ),
        None,
    )


def _passed_command(args: List[str]) -> bool:
    return len(args) > 0


def _get_command_by_argument(command: str, template: Dict):
    commands = _get_commands(template)

    if command not in commands:
        raise CommandNotFoundException(command)
    else:
        return _get_command_by_name(command, template)


def _passed_action(args: List[str]) -> bool:
    return len(args) > 1


def _get_action_by_argument(action: str, command: Dict):
    actions = _get_actions(command)
    # actions_alias = _get_actions_alias(command)

    if action not in actions:
        raise ActionNotFoundException(action)
    else:
        return _get_action_by_name(action, command)


def _get_command(args, template):
    if _passed_command(args):
        command = _get_command_by_argument(args[0], template)

        if len(args) > 1 and "-h" == args[1]:
            raise ShowHelpException(command.get("help"))

        return command
    else:
        return _get_command_by_template(template)


def _get_action(args, command):
    if _passed_action(args):
        action = _get_action_by_argument(args[1], command)

        if len(args) > 2 and "-h" == args[2]:
            raise ShowHelpException(action.get("help"))

        return action
    else:
        return _get_action_by_command(command)


@click.command(
    context_settings=dict(
        allow_extra_args=True,
        ignore_unknown_options=True,
    )
)
@click.version_option(
    package_name="git-phoenix",
    prog_name="Git Phoenix",
)
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def main(args):
    try:
        require_git_repo()
        template = _get_template()

        try:
            validate(instance=template, schema=PHOENIX_SCHEMA)

            command = _get_command(args=args, template=template)
            action = _get_action(args=args, command=command)
            variables = template.get("variables")
            arguments = args[2:]

            # Create execution object to carry template and
            # arguments through execution
            execution = Execution(
                command=command,
                action=action,
                variables=variables,
                arguments=arguments,
            )

            print("Firing rules...")

            fire_rules(execution)
        except ValidationError as e:
            raise InvalidTemplateException(e.message)
    except PhoenixException as e:
        print(e.message)
    except Exception as e:
        print(str(e))



import git
from git import GitCommandError, Repo

from src.core.exceptions import (
    NotAGitRepoException,
    BranchAlreadyExistsException,
)

GIT_REPO: Repo


def get_config(key):
    try:
        global GIT_REPO
        return GIT_REPO.git.config(key)
    except GitCommandError:
        return f"Config {key} not found"


def get_tags():
    global GIT_REPO
    return [
        version.replace("v", "") for version in GIT_REPO.git.tag().split("\n")
    ]


def require_git_repo():
    if not _is_git_repo():
        raise NotAGitRepoException()


def _is_git_repo():
    try:
        global GIT_REPO
        GIT_REPO = Repo(".", search_parent_directories=True)
        return True
    except git.exc.InvalidGitRepositoryError:
        return False


def has_unstaged_files():
    unstaged_files = GIT_REPO.git.status("--porcelain")

    if not unstaged_files:
        return False

    return True


def is_ahead():
    try:
        branch = retrieve_current_branch()

        commits_ahead = GIT_REPO.iter_commits(
            "source/" + branch + ".." + branch
        )
        number_of_commits = sum(1 for c in commits_ahead)

        return number_of_commits > 0
    except GitCommandError:
        print(
            f"Your current branch ({branch}) doesn't exists on remote "
            f"repo. Please use git push source {branch}."
        )


def retrieve_current_branch():
    return GIT_REPO.active_branch.name


def checkout(branch):
    # Logger.info(
    #     self=GitCommons,
    #     msg=(
    #         "Checking out branch"
    #         + PythonCommons.LIGHT_CYAN
    #         + " {}"
    #         + PythonCommons.NC
    #         + "..."
    #     ).format(branch),
    # )

    GIT_REPO.git.checkout(branch)


def _validate_existence(branch):
    global GIT_REPO
    remotes = [branch.name.replace("origin/", "") for branch in GIT_REPO.refs]
    local = GIT_REPO.branches

    if branch in remotes or branch in local:
        raise BranchAlreadyExistsException()


def checkout_new_branch(source, branch):
    checkout(source)
    pull(source)
    # Logger.info(
    #     self=GitCommons,
    #     msg=(
    #         "Creating branch"
    #         + PythonCommons.LIGHT_CYAN
    #         + " {} "
    #         + PythonCommons.NC
    #         + "based on"
    #         + PythonCommons.LIGHT_CYAN
    #         + " {}"
    #         + PythonCommons.NC
    #         + "..."
    #     ).format(branch, source),
    # )
    try:
        _validate_existence(branch)

        GIT_REPO.git.checkout(source, b=branch)
    except GitCommandError:
        pass
        # Logger.error(
        #     self=GitCommons,
        #     msg=(
        #         "Branch"
        #         + PythonCommons.LIGHT_CYAN
        #         + " {} "
        #         + PythonCommons.NC
        #         + "already exists!"
        #     ).format(branch),
        # )


def merge(branch, destination, allow_merge_again):
    git = Repo(".", search_parent_directories=True).git

    if allow_merge_again or not _already_merged(destination, branch):
        checkout(destination)
        pull(destination)
        Logger.info(
            cls=GitCommons,
            msg=(
                "Merging branch"
                + PythonCommons.LIGHT_CYAN
                + " {} "
                + PythonCommons.NC
                + "into"
                + PythonCommons.LIGHT_CYAN
                + " {}"
                + PythonCommons.NC
                + "..."
            ).format(branch, destination),
        )

        try:
            git.merge(branch, "--no-ff")
        except:
            Logger.warn(
                cls=GitCommons,
                msg="You have conflicts on your working tree! Resolve them before commiting!",
            )
            raise ExecutionException()
    else:
        Logger.warn(
            cls=GitCommons,
            msg=(
                "Branch"
                + PythonCommons.LIGHT_CYAN
                + " {} "
                + PythonCommons.NC
                + "already merged into"
                + PythonCommons.LIGHT_CYAN
                + " {}"
                + PythonCommons.NC
                + "!"
            ).format(branch, destination),
        )


def pull(branch):
    # Logger.info(
    #     self=GitCommons,
    #     msg=(
    #         "Updating branch "
    #         + PythonCommons.LIGHT_CYAN
    #         + "{}"
    #         + PythonCommons.NC
    #         + "..."
    #     ).format(branch),
    # )
    try:
        GIT_REPO.git.pull()
    except GitCommandError:
        pass


def delete(pattern):
    git = Repo(".", search_parent_directories=True).git
    Logger.info(
        cls=GitCommons,
        msg=(
            "Deleting branch(es) "
            + PythonCommons.LIGHT_CYAN
            + "{}"
            + PythonCommons.NC
            + "..."
        ).format(pattern),
    )

    branches = [
        branch.replace(" ", "") for branch in git.branch().splitlines()
    ]
    branches = list(filter(lambda x: x.startswith(pattern), branches))
    [git.execute(["git", "branch", "-D", branch]) for branch in branches]


def _already_merged(destination, branch):
    merged_branches = log(
        ["--oneline", "--merges", "--grep=into", destination]
    ).splitlines()

    for merged_branch in merged_branches:
        translated_merge_message = _translate_merge_message(msg=merged_branch)

        if translated_merge_message["source"] == branch:
            return True

    return False


def log(parameters):
    git = Repo(".", search_parent_directories=True).git

    if parameters:
        return git.log(parameters)

    return git.log()


def _translate_merge_message(msg):
    translated_merge_message = {}

    origin = msg.split("branch ")

    if origin[1:]:
        origin = origin[1]
    else:
        origin = origin[0]

    origin = origin[: origin.index(" ")]

    translated_merge_message["source"] = origin.replace("'", "")
    translated_merge_message["destination"] = msg.split("into ")[-1].replace(
        "'", ""
    )

    return translated_merge_message



import re
import ast

from src.core.phoenix import define_pattern, read_input, determine_pattern
from src.core.actions.executable import Executable
from src.core.logger import Logger
from src.core import merge


class Merge(Executable):
    def __init__(self, execution, action_execution):
        super().__init__(execution, action_execution)

    def execute(self):
        if hasattr(self, "origin_pattern"):
            pattern = re.compile(self.origin_pattern)

            if not pattern.match(self.origin):
                pattern_msg = None

                if hasattr(self, "origin_pattern_example"):
                    pattern_msg = self.origin_pattern_example
                else:
                    pattern_msg = self.origin_pattern

                Logger.warn(
                    cls=Merge,
                    msg=(
                        "You are merging branch"
                        + PythonCommons.LIGHT_CYAN
                        + " {}"
                        + PythonCommons.NC
                        + ". Merge a branch with this name pattern: "
                        + PythonCommons.LIGHT_CYAN
                        + "{}"
                        + PythonCommons.NC
                    ).format(self.origin, pattern_msg),
                )
                Logger.error(
                    cls=Merge,
                    msg="Invalid source! Please execute the command with a valid source branch!",
                )

        all_destinations = self.destination[:]

        for destination in self.destination:
            try:
                all_destinations.pop(0)
                merge(self.origin, destination, self.allow_merge_again)
            except ExecutionException:
                if len(all_destinations) > 0:
                    Logger.warn(
                        cls=Merge,
                        msg=(
                            "Due to the conflict, the branch"
                            + PythonCommons.LIGHT_CYAN
                            + " {} "
                            + PythonCommons.NC
                            + "couldn't be merged with branch(es)"
                            + PythonCommons.LIGHT_CYAN
                            + " {}"
                            + PythonCommons.NC
                            + "!"
                        ).format(self.origin, ", ".join(all_destinations)),
                    )

                raise ExecutionException()

    def _parse(self):
        if not hasattr(self, "source"):
            if [] == self.execution.args[0:]:
                self.origin = read_input(
                    cls=Merge, msg="Inform the source branch name:"
                )
            else:
                self.origin = self.execution.args[0]

        if hasattr(self, "origin_pattern"):
            self.origin_pattern = determine_pattern(self.origin_pattern)

        if not hasattr(self, "destination"):
            if not hasattr(self, "destination_strategy"):
                if [] == self.execution.args[1:]:
                    self.destination = read_input(
                        cls=Merge, msg="Inform the destination branch(es):"
                    ).split(" ")
                else:
                    self.destination = [self.execution.args[1]]
            else:
                strategy = self.destination_strategy["strategy"]

                if strategy == "search_pattern":
                    pattern = self.destination_strategy["pattern"]

                    for i in range(len(pattern)):
                        pattern[i] = self._change_value(pattern[i])

                    index_to_search = self.destination_strategy["index"]
                    pattern_to_search = define_pattern(
                        pattern[index_to_search], False, False
                    )

                    pattern_found = re.search(
                        pattern_to_search, self.origin
                    ).group(0)
                    pattern[index_to_search] = pattern_found

                    self.destination = [
                        PhoenixCommons.determine_prefix(pattern)
                    ]
                else:
                    Logger.error(
                        cls=Merge,
                        msg=(
                            "Strategy"
                            + PythonCommons.LIGHT_PURPLE
                            + " {} "
                            + PythonCommons.NC
                            + "not implemented yet!"
                        ).format(strategy),
                    )
        if hasattr(self, "allow_merge_again"):
            self.allow_merge_again = self.allow_merge_again
        else:
            self.allow_merge_again = False

    def confirm_execution(self):
        destination = ", ".join(self.destination)

        return super()._confirm_execution(
            cls=Merge,
            msg=(
                "Confirm merging branch"
                + PythonCommons.LIGHT_CYAN
                + " {} "
                + PythonCommons.NC
                + "into"
                + PythonCommons.LIGHT_CYAN
                + " {}"
                + PythonCommons.NC
                + "?"
            ).format(self.origin, destination),
        )


class MergeRequest(Executable):
    pass



import regex
import sys
from src.core.actions.executable import Executable
from src.core.logger import Logger


class DeleteBranch(Executable):

    def __init__(self, execution, action_execution):
        super().__init__(execution, action_execution)

    def execute(self):
        delete(self.pattern)

    def _parse(self):
        if hasattr(self, "pattern"):
            self.pattern = "".join(self.pattern)
        else:
            self.pattern = read_input(cls=DeleteBranch, msg="Inform the branch pattern to delete:")

    def confirm_execution(self):
        return super()._confirm_execution(cls=DeleteBranch, msg=("Confirm deleting branch(es)" +
                                                                 PythonCommons.LIGHT_CYAN +
                                                                 " {}" +
                                                                 PythonCommons.NC +
                                                                 "?").format(self.pattern))
